/* consoletunes.c - play notes on the internal speaker of a linux
 * box so the machine can signal me
 *
 * Copyright (c) 1997 Tony Aiuto
 *
 * To report problems with this tool or to suggest enhancements send mail
 * to tony@aiu.to
 *
 */

#ifndef lint
static char _copyright[] = "@(#) Copyright (c) 1997, Tony Aiuto";
#endif

/*
 * Author: Tony Aiuto (tony@aiu.to)
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor their employers are responsible for
 *    any consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors must appear
 *    in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This copyright notice may not be removed or altered.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL TONY AIUTO, HIS CURRENT OR PAST EMPLOYERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * Edit History:
 * $Log: consolenotes.c,v $
 */

#ifndef lint
static char rcsid[] = "@(#) $Id$";
#endif


#define	TRUE_PITCH	1

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#include <asm/timex.h>
#include <math.h>

extern	int	optind;
extern	char	*optarg;


void usage()
{
	fprintf(stderr, "Warning: need to be root to work\n");
	fprintf(stderr, "usage: consolenotes note length ....\n");
	fprintf(stderr, "\tnotes of form c{n} where n=0 yields middle C\n");
	fprintf(stderr, "\tlength of 1 = whole note, 2 = half note ...\n");
	fprintf(stderr, "\texample:\n\tconsolenotes c-1 4 d-1 4 e-1 4 f-1 4 g-1 4 a-1 4 b-1 4 \\\n\tc0 4 d0 4 e0 4 f0 4 g0 4 a0 4 b0 4 \\\n\tc1 4 d1 4 e1 4 f1 4 g1 4 a1 4 b1 4 c2 4\n");
}

void tone(float pitch, int duration)
{
	static	FILE *out = NULL;

	if(out == NULL) {
		out = fopen("/dev/console", "w");
	}
	if(out == NULL) return;

#if BELL_CHAR_METHOD
	if(duration <= 0) {
		fprintf(out, "");
	} else {
		int	bells = 8 / duration;
		if(bells < 1) bells = 1;
		while(bells-- > 0) {
			fprintf(out, "");
		}
	}
	fflush(out);
#endif

/*	KDMKTONE
              Generate  tone  of  specified length.  The lower 16
              bits of argp specify the period  in  clock  cycles,
              and  the  upper  16 bits give the duration in msec.
              If the duration is zero, the sound is  turned  off.
              Control  returns  immediately.  For example, argp =
              (125<<16) + 0x637 would specify the  beep  normally
              associated with a ctrl-G.


	KIOCSOUND
              Start  or stop sound generation.  The lower 16 bits
              of argp specify the period in  clock  cycles  (that
              is,  argp  =  1193180/frequency).   argp  = 0 turns
              sound off.  In either case, control returns immedi
              ately.
*/

#if OLD
	ioctl(fileno(out), KDMKTONE,  (1000 / duration) << 16 |
						(int)(1193180 / pitch));
#else
	ioctl(fileno(out), KDMKTONE,  (1000 / duration) << 16 |
					(int)(CLOCK_TICK_RATE / pitch));

#endif

	return;
}

#define	NOTE_A	440

float getNote(const char *s)
{
	int	i;
	int	lname;
	const char	*tmp_s;

	static struct {
		char	*name;
		float	pitch;
	} notes[] = {

#if TRUE_PITCH
		/* pitches of middle octaves */
		{"c",  261.6255653006},
		{"c#", 277.1826309769},
		{"d",  293.6647679174}, 
		{"db", 277.1826309769},
		{"d#", 311.1269837221},
		{"e",  329.6275569129},
		{"eb", 311.1269837221},
		{"f",  349.2282314330},
		{"f#", 369.9944227116},
		{"g",  391.9954359817},
		{"gb", 369.9944227116},
		{"g#", 415.3046975799},
		{"a",  440.0000000000},
		{"ab", 415.3046975799},
		{"a#", 466.1637615181},
		{"b",  493.8833012561},
		{"bb", 466.1637615181},

#else
		/* Single char note names last */
		{"c", 261.0},
		{"d", 293.625},
		{"e", 326.250},
		{"f", 348.0},
		{"g", 391.5},
		{"a", 435.0},		/* I thought it is 440 */
		{"b", 489.375}
#endif
	};

	/* guard againt bad input */
	if(s == NULL || *s == '\0') return 0;

	/* look for midi note number syntax */
	if(*s == 'm' && isdigit(s[1])) {
		int midiNote = atoi(&s[1]);

		return NOTE_A / 32 * pow(2.0, ((double)(midiNote-8))/12.0);
	}

	/* figure out length of note part without octave modifier */
	lname = 0;
	tmp_s = s; 
	while(*tmp_s && !isdigit(*tmp_s) && *tmp_s != '-') {
		lname++;
		tmp_s++;
	}
	if(lname <= 0) return 0;

	/* printf("testing '%s', lname=%d\n", s, lname); */
	for(i = 0; i < sizeof(notes) / sizeof(notes[0]); i++) {
		/* printf("... '%s'\n", notes[i].name); */
		if(strcmp(s, notes[i].name) == 0) return notes[i].pitch;

		/* Test for forms, a0, a1, a-1, ... and compute pitch */
		if(strncmp(s, notes[i].name, lname) == 0) {
			int octave = atoi(&s[lname]);
			float ret = notes[i].pitch;
			if(octave > 0) {
				while(octave-- > 0) {
					ret *= 2;
				}
			} else if(octave < 0) {
				while(octave++ < 0) {
					ret /= 2;
				}
			}
			return ret;
		}
	}
	return 0;		/* signal no note lookup */
}

main(int argc, char *argv[])
{
	int	c;
	int	verbose = 0;

	while((c = getopt(argc, argv, "v")) != EOF) {
		switch(c) {
		case 'v':
			verbose = 1;
			break;
		default:
			usage();
			exit(1);
		}
	}
	if(optind >= argc) {
		tone(440, 4);
	} else {
		int	pitch = 440;
		for(; optind < argc; optind++) {
			char *arg = argv[optind];

			int duration;
			float	newpitch;

			newpitch = getNote(arg);
			/* if the note changes, we just save that and
			 * loop again to pick up the duration
			 */
			if(newpitch > 0) {
				pitch = newpitch;
				continue;
			}

			/* if the note is not found, we try to interpret
			 * the arg as a duration
			 */

			duration = atoi(arg);
			if(duration < 0) {
				pitch = (int) -duration;
				continue;
			}
			if(duration > 0) {
				if(verbose) {
					printf("%d %d\n", pitch, duration);
				}
				tone(pitch, duration);
				usleep( 1100000 / duration);
			}
		}
	}
	exit(0);
}
